import 'dart:convert';

import 'package:crypto/crypto.dart';
import 'package:http/http.dart' as http;

const String ak = '/** Provide your AK **/';
const String sk = '/** Provide your SK **/';
// const String token = '/** Provide your Token **/';
const String endpoint = '/** Provide your endpoint **/';
const String bucket = '/** Provide your bucket name **/';
const String object = '/** Provide your object key **/';
const String region = '/** Provide your region **/';

const String service = 'tos';
const String request = 'request';
const String algorithm = 'TOS4-HMAC-SHA256';

void main() {
  // listBuckets();
  // putObject();
  // listObjects();
  // getObject();
  // deleteObject();
  // multipartUpload();
}

// 列举桶
void listBuckets() async {
  var requestUrl = Uri.parse(endpoint);
  var headers = getAuthorization(requestUrl, 'GET');
  try {
    var resp = await http.get(requestUrl, headers: headers);
    print('Response status: ${resp.statusCode}');
    var body = jsonDecode(resp.body);
    if (resp.statusCode >= 300) {
      printError(body);
    } else {
      for (var bucket in body['Buckets']) {
        print(bucket['Name']);
      }
    }
  } catch (err) {
    print(err);
  }
}

// 上传对象
void putObject() async {
  var requestUrl = getBucketRequestUrl(objectKey: object);
  var headers = getAuthorization(requestUrl, 'PUT');
  try {
    var resp =
        await http.put(requestUrl, headers: headers, body: 'hello world');
    print('Response status: ${resp.statusCode}');
    if (resp.statusCode >= 300) {
      var body = jsonDecode(resp.body);
      printError(body);
    } else {
      print('putObject succeed');
    }
  } catch (err) {
    print(err);
  }
}

// 列举对象
void listObjects() async {
  var requestUrl = getBucketRequestUrl(query: 'max-keys=5');
  var headers = getAuthorization(requestUrl, 'GET');
  try {
    var resp = await http.get(requestUrl, headers: headers);
    print('Response status: ${resp.statusCode}');
    var body = jsonDecode(resp.body);
    if (resp.statusCode >= 300) {
      printError(body);
    } else {
      for (var content in body['Contents']) {
        print(content['Key']);
      }
    }
  } catch (err) {
    print(err);
  }
}

// 下载对象
void getObject() async {
  var requestUrl = getBucketRequestUrl(objectKey: object);
  var headers = getAuthorization(requestUrl, 'GET');
  try {
    var resp = await http.get(requestUrl, headers: headers);
    print('Response status: ${resp.statusCode}');
    if (resp.statusCode >= 300) {
      var body = jsonDecode(resp.body);
      printError(body);
    } else {
      print('getObject succeed, body: ${resp.body}');
    }
  } catch (err) {
    print(err);
  }
}

// 删除对象
void deleteObject() async {
  var requestUrl = getBucketRequestUrl(objectKey: object);
  var headers = getAuthorization(requestUrl, 'DELETE');
  try {
    var resp = await http.delete(requestUrl, headers: headers);
    print('Response status: ${resp.statusCode}');
    if (resp.statusCode >= 300) {
      var body = jsonDecode(resp.body);
      printError(body);
    } else {
      print('deleteObject succeed');
    }
  } catch (err) {
    print(err);
  }
}

// 分片上传
void multipartUpload() async {
  var requestUrl = getBucketRequestUrl(objectKey: object, query: 'uploads');
  var headers = getAuthorization(requestUrl, 'POST');
  try {
    // 1. 初始化分段上传任务
    var resp = await http.post(requestUrl, headers: headers);
    print('Response status: ${resp.statusCode}');
    var body = jsonDecode(resp.body);
    if (resp.statusCode >= 300) {
      printError(body);
    } else {
      var uploadId = body['UploadId'];
      print('init upload succeed, uploadID: ${uploadId}');

      // 2. 上传分段
      String? etag = '';
      requestUrl = getBucketRequestUrl(
          objectKey: object, query: 'partNumber=1&uploadId=${uploadId}');
      headers = getAuthorization(requestUrl, 'PUT');
      resp = await http.put(requestUrl, headers: headers, body: 'hi world');
      print('Response status: ${resp.statusCode}');
      if (resp.statusCode >= 300) {
        body = jsonDecode(resp.body);
        printError(body);
      } else {
        etag = resp.headers['etag'];
        print('upload part succeed, etag: ${etag}');
      }

      if (etag != null) {
        // 3. 合并段
        requestUrl = getBucketRequestUrl(
            objectKey: object, query: 'uploadId=${uploadId}');
        headers = getAuthorization(requestUrl, 'POST');
        var mergeBody = jsonEncode({
          'Parts': [
            {'PartNumber': 1, 'ETag': etag},
          ]
        });
        resp = await http.post(requestUrl, headers: headers, body: mergeBody);
        print('Response status: ${resp.statusCode}');
        body = jsonDecode(resp.body);
        if (resp.statusCode >= 300) {
          printError(body);
        } else {
          print('complete upload succeed, location: ${body['Location']}');
        }
      }
    }
  } catch (err) {
    print(err);
  }
}

// Authorization: TOS4-HMAC-SHA256 Credential={AccessKeyId}/{CredentialScope}, SignedHeaders={SignedHeaders}, Signature={Signature}
Map<String, String> getAuthorization(Uri requestUrl, String method) {
  var utc = getUtcNow();
  var longDate = getLongTime(utc);
  var headers = Map<String, String>();
  headers['Host'] = requestUrl.host;
  headers['Date'] = longDate;
  // support sts token if necessary
  // headers['x-tos-security-token'] = token;
  var list = getCanonicalHeadersAndSignedHeaders(headers);
  var canonicalHeaders = list[0];
  var signedHeaders = list[1];
  var shortDate = getShortTime(utc);

  var credentialScope = getCredentialScope(shortDate);
  var canonicalUri = requestUrl.path.trim();
  if (canonicalUri.isEmpty) {
    canonicalUri = '/';
  }
  canonicalUri = uriEncode(canonicalUri, true);

  var queryMap = Uri.splitQueryString(requestUrl.query);
  var queryList = [];

  for (var entry in queryMap.entries) {
    queryList.add(entry.key);
  }
  queryList.sort();

  var canonicalQueryString = '';
  var index = 0;
  for (var key in queryList) {
    canonicalQueryString += '${uriEncode(key, false)}=${queryMap[key]?.trim()}';
    if (index != queryList.length - 1) {
      canonicalQueryString += '&';
    }
    index++;
  }

  var canonicalRequest = getCanonicalRequest(method, canonicalUri,
      canonicalQueryString, canonicalHeaders, signedHeaders);

  // print('canonicalRequest:\n${canonicalRequest}');

  var authorization =
      '${algorithm} Credential=${ak}/${credentialScope}, SignedHeaders=${signedHeaders}, Signature=${getSignature(shortDate, longDate, credentialScope, canonicalRequest)}';
  headers['Authorization'] = authorization;
  return headers;
}

String getCredentialScope(String shortDate) {
  return '${shortDate}/${region}/${service}/${request}';
}

List<String> getCanonicalHeadersAndSignedHeaders(Map<String, String> headers) {
  var signedHeadersList = [];
  for (var entry in headers.entries) {
    var key = entry.key;
    var _key = key.toLowerCase();
    if (_key == 'host' || _key.startsWith('x-tos')) {
      signedHeadersList.add(key);
    }
  }
  signedHeadersList.sort();

  var canonicalHeaders = '';
  var signedHeaders = '';
  var index = 0;
  for (String key in signedHeadersList) {
    var _key = key.toLowerCase();
    canonicalHeaders += '${_key}:${headers[key]?.trim()}\n';
    signedHeaders += _key;
    if (index != signedHeadersList.length - 1) {
      signedHeaders += ';';
    }
    index++;
  }
  return [canonicalHeaders, signedHeaders];
}

// Signature = HexEncode(HMAC(kSigning, StringToSign))
String getSignature(String shortDate, longDate, String credentialScope,
    String canonicalRequest) {
  return '${Hmac(sha256, getSigningKey(shortDate)).convert(utf8.encode(getStringToSign(longDate, credentialScope, canonicalRequest)))}';
}

// kSecret = *Your Secret Access Key*
// kDate = HMAC(kSecret, Date)
// kRegion = HMAC(kDate, Region)
// kService = HMAC(kRegion, Service)
// kSigning = HMAC(kService, 'request')
List<int> getSigningKey(String shortDate) {
  var kDate =
      Hmac(sha256, utf8.encode(sk)).convert(utf8.encode(shortDate)).bytes;
  var kRegion = Hmac(sha256, kDate).convert(utf8.encode(region)).bytes;
  var kService = Hmac(sha256, kRegion).convert(utf8.encode(service)).bytes;
  var kSigning = Hmac(sha256, kService).convert(utf8.encode(request)).bytes;
  return kSigning;
}

// StringToSign = Algorithm + '\n' + RequestDate + '\n' + CredentialScope + '\n' + HexEncode(Hash(CanonicalRequest))
String getStringToSign(
    String longDate, credentialScope, String canonicalRequest) {
  return '${algorithm}\n${longDate}\n${credentialScope}\n${sha256.convert(utf8.encode(canonicalRequest))}';
}

// CanonicalRequest = HTTPRequestMethod + '\n' + CanonicalURI + '\n' + CanonicalQueryString + '\n' + CanonicalHeaders + '\n' + SignedHeaders + '\n' + HexEncode(Hash(RequestPayload))
String getCanonicalRequest(String method, canonicalUri, canonicalQueryString,
    canonicalHeaders, signedHeaders) {
  return '${method.toUpperCase()}\n${canonicalUri}\n${canonicalQueryString}\n${canonicalHeaders}\n${signedHeaders}\n${sha256.convert(utf8.encode(''))}';
}

DateTime getUtcNow() {
  return new DateTime.now().toUtc();
}

String getShortTime(DateTime utc) {
  return '${utc.year.toString()}${utc.month.toString().padLeft(2, '0')}${utc.day.toString().padLeft(2, '0')}';
}

String getLongTime(DateTime utc) {
  return '${getShortTime(utc)}T${utc.hour.toString().padLeft(2, '0')}${utc.minute.toString().padLeft(2, '0')}${utc.second.toString().padLeft(2, '0')}Z';
}

void printError(dynamic body) {
  print('Code: ${body['Code']}');
  print('Message: ${body['Message']}');
  print('RequestId: ${body['RequestId']}');
  print('canonical_request:\n${body['canonical_request']}');
  print('');
  print('StringToSign:\n${body['StringToSign']}');
}

Uri getBucketRequestUrl({String? objectKey, String? query}) {
  var uri = Uri.parse(endpoint);
  var base = uri.scheme + '://' + bucket + '.' + uri.host;
  if (objectKey != null && objectKey.trim() != '') {
    base += '/' + objectKey.trim();
  }

  if (query != null && query.trim() != '') {
    base += '?' + query;
  }
  // print(base);
  return Uri.parse(base);
}

String uriEncode(String input, bool escapeSlash) {
  var output = Uri.encodeComponent(input);

  if (escapeSlash) {
    output = output.replaceAll('%2F', '/');
  }
  output = output.replaceAll('*', '%2A');
  output = output.replaceAll('+', '%20');
  output = output.replaceAll('~', '%7E');
  output = output.replaceAll('%2520', '%20');
  return output;
}
